"""
QMTest extension module for Moore and related tests.

- compareTimingTableEvents, extract the nubmer of events from the timing table
                            along with the alg name, and use that to get
                            a comparison versus some references
                            Categorize and pretty print the results.

"""
import re
import difflib
from math import fabs
from GaudiTesting.BaseTest import BaseTest
from . import timingtable


def extractEventsFromTiming(stdout,beginswith=None):
    """
    extract the timing table from the stdout
    returns a tuple, first a list of Alg:events, then a dict of {alg : event}
    """
    timing=[]
    timingdict={}
    for line in stdout.split('\n'):
        if 'TIMER' not in line:
            continue
        if beginswith is not None and beginswith not in line:
            continue
        line=line.replace('\t',' ')
        line=line.replace('  ',' ')
        line=line[line.find('INFO')+4:]
        line=line.strip().rstrip()
        line=line.split('|')
        if len(line)<5:
            continue
        if not line[0].startswith(beginswith):
            continue
        name=line[0].strip().rstrip().lstrip()
        num=line[4].strip().rstrip().lstrip()
        timing.append(': '.join([name,num]))
        timingdict[name]=int(num)
    return timing, timingdict

#function to add results
def listToResult(testname,resname,thelist,refresultdict,timingdict,toomany,resstr=None,result=None):
    """
    converts a list into a string for adding into a result.
    testname: the name of this test, used for the result name
    resname: the name of this specific result, the result name is made from
              testname+resname.lower()
    thelist: list of whatever to make into a quote
    refresultsdict: dictionary of {referenceAlg : #events}
    timingdict: dictionary of {testAlg : #events}
    toomany: flag to allow the full list as a quote, or only the first 10 entries
    resstr: override the resname in the quote, but not the results entry,
            usually the results entry starts with resname.split('_')[-1]
    """
    #global result
    #print "64"
    if not len(thelist):
        return
    #print "67"
    if resstr is None:
        resstr=resname.split('_')[-1]
    prstr=resstr+": "+str(len(thelist))
    #print "71"
    if len(thelist)>10 and toomany:
        thelist=thelist[:10]
    #print "74"
    newlist=[]
    for l in thelist:
        if l in refresultdict and l in timingdict:
            newlist.append(l+' '+str(refresultdict[l])+'->'+str(timingdict[l]))
        elif l in refresultdict:
            newlist.append(l+' '+str(refresultdict[l])+'->?')
        elif l in timingdict:
            newlist.append(l+' ?->'+str(timingdict[l]))
        else:
            newlist.append(l)
    prstr=prstr+'\n\t'+',\n\t'.join(newlist)
    #print "76"
    result[testname+'.'+resname]=result.Quote(prstr)
    #print "78"
    return

def eventTableRef(filename,beginswith=None):
    """
    Reads a timing table ref, returns a tuple:
         firstly list of [alg: num]
         then a dict of {alg:num}
    """
    reftiming=open(filename)
    refresult=reftiming.read()
    reftiming.close()
    refresult.replace('  ',' ')#ignore multiple whitespace
    refresult=refresult.split('\n')
    refdict={}
    refresult=[ref.strip().rstrip().lstrip() for ref in refresult]
    refresult=[ref for ref in refresult if len(ref)>2]#ignore white lines
    if beginswith is not None:
        refresult=[ref for ref in refresult if ref.startswith(beginswith)]
    for ref in refresult:
        refdict[':'.join(ref.split(':')[:-1])]=int(ref.split(':')[-1])
    return refresult, refdict

def compareTimingTableEvents(testname,result,causes,stdout,refFile,beginswith=None,max=200,extranewref=''):
    """
    compare the timing table with a test timing table
    testname: the name of this qmtest
    result: the result to add to
    causes: the causes to add to
    stdout: the stdout, a string
    refFile: where to find the reference, a string
    beginswith: ignore timing table entries which do not start with this
    max: how many differences would be too many to print? default 200
    extranewref: whether to append anything extra to the newly created ref file
    """
    #parse the tables
    refresult,refresultdict=eventTableRef(refFile,beginswith)
    timing, timingdict=extractEventsFromTiming(stdout,beginswith)

    #compare, write out new refs?
    if refresult==timing or refresultdict==timingdict:
        return
    causes.append('#events in timing tables disagree')
    if len(extranewref) and not extranewref.startswith('.'):
        extranewref='.'+extranewref
    if beginswith is not None:
        extranewref=extranewref+'.'+beginswith
    f=open(refFile+extranewref+'.new','w')
    f.write('\n'.join(timing))
    f.close()
    #categorize the differences
    missing=[r for r in refresultdict if r not in timingdict]
    added=[r for r in timingdict if r not in refresultdict]
    diff=[r for r in refresultdict if ((r in timingdict) and (timingdict[r]!=refresultdict[r]))]
    singleEventDiffs=[r for r in diff if abs(refresultdict[r]-timingdict[r])<2]
    lineFailureDiff=[r for r in diff if ((r not in singleEventDiffs) and (refresultdict[r]!=0) and (timingdict[r]==0))]
    tenPercentDiffs=[r for r in diff if ((r not in singleEventDiffs) and (r not in lineFailureDiff) and (fabs(float(refresultdict[r]-timingdict[r])/float(refresultdict[r]+timingdict[r]))<0.1))]
    rest=[r for r in diff if ((r not in singleEventDiffs) and (r not in lineFailureDiff) and (r not in tenPercentDiffs))]

    #get the results
    toomany=False
    if len(missing)+len(added)+len(diff)>max:
        result[testname+'.a_total']=result.Quote('More than 200 differences, summary only: ' +str(len(missing)+len(added)+len(diff)))
        toomany=True
    else:
        result[testname+'.a_total']=result.Quote('Total: ' +str(len(missing)+len(added)+len(diff)))
    listToResult(testname,"b_Missing",missing,refresultdict,timingdict,toomany,result=result)
    listToResult(testname,"c_Added",added,refresultdict,timingdict,toomany,result=result)
    listToResult(testname,"d_Single",singleEventDiffs,refresultdict,timingdict,toomany,"+/- 1 event",result=result)
    listToResult(testname,"e_failure",lineFailureDiff,refresultdict,timingdict,toomany,"zero events",result=result)
    listToResult(testname,"f_10pc",tenPercentDiffs,refresultdict,timingdict,toomany,"+/- <10 %",result=result)
    listToResult(testname,"g_Rest",rest,refresultdict,timingdict,toomany,result=result)


def compareTimingTableFiles(testname,result,causes,myfile,refFile,beginswith=None,max=200,extranewref='',timingthreshold=10.):
    """
    compare the timing table with a test timing table
    testname: the name of this qmtest
    result: the result to add to
    causes: the causes to add to
    stdout: the stdout, a string
    refFile: where to find the reference, a string
    beginswith: ignore timing table entries which do not start with this
    max: how many differences would be too many to print? default 200
    extranewref: whether to append anything extra to the newly created ref file
    """
    from LHCbAlgs import TTParser
    #parse the tables
    import os
    if not os.path.exists(myfile):
        causes.append("Cannot compare timing, timing table not created "+myfile)
        return
    
    if not os.path.exists(refFile):
        causes.append("Cannot compare timing, reference table does not exit "+refFile)
        return
    
    TTref=TTParser.parse(refFile)
    TTthis=TTParser.parse(myfile)
    #
    missing,different,added=TTParser.compare(TTref,TTthis,threshold=timingthreshold)
    if beginswith:
        missing=[m for m in missing if m.startswith(beginswith)]
        different=[d for d in different if d.startswith(beginswith)]
        added=[a for a in added if a.startswith(beginswith)]
    if len(missing+different+added)==0:
        return
    
    #work out if it was the number of events or the timing that caused this
    refdict=TTParser.event_av_dict(TTref)
    thisdict=TTParser.event_av_dict(TTthis)
    events=[]
    timings=[]
    for d in different:
        if refdict[d][0]!=thisdict[d][0]:
            events.append(d)
        elif refdict[d][1]<thisdict[d][1] and refdict[d][1]>0.0001:
            #allow faster, but not slower, and ignore <10 events for noise
            if refdict[d][0]>10:
                timings.append(d)
    
    different=events+timings
    if len(events):
        causes.append('#timing tables disagree')
        
        if len(extranewref) and not extranewref.startswith('.'):
            extranewref='.'+extranewref
        if beginswith is not None:
            extranewref=extranewref+'.'+beginswith
        import os
        os.system("cp "+myfile+" "+refFile+extranewref+'.new')
    
    #categorize the differences
    singleEventDiffs=[r for r in events if abs(refdict[r][0]-thisdict[r][0])<2]
    lineFailureDiff=[r for r in events if ((r not in singleEventDiffs) and (refdict[r][0]!=0) and (thisdict[r][0]==0))]
    tenPercentDiffs=[r for r in events if ((r not in singleEventDiffs) and (r not in lineFailureDiff) and (fabs(float(refdict[r][0]-thisdict[r][0])/float(refdict[r][0]+thisdict[r][0]))<0.1))]
    rest=[r for r in events if ((r not in singleEventDiffs) and (r not in lineFailureDiff) and (r not in tenPercentDiffs))]
    
    #get the results
    toomany=False
    if len(missing)+len(added)+len(different)>max:
        result[testname+'.a_total']=result.Quote('More than 200 differences, summary only: ' +str(len(missing)+len(added)+len(different)))
        toomany=True
    else:
        result[testname+'.a_total']=result.Quote('Total: ' +str(len(missing)+len(added)+len(different)))
    refresultdict={}
    timingdict={}
    for r,v in refdict.iteritems():
        refresultdict[r]=v[0]
    for r,v in thisdict.iteritems():
        timingdict[r]=v[0]
    
    listToResult(testname,"b_Missing",missing,refresultdict,timingdict,toomany,result=result)
    listToResult(testname,"c_Added",added,refresultdict,timingdict,toomany,result=result)
    listToResult(testname,"d_Single",singleEventDiffs,refresultdict,timingdict,toomany,"+/- 1 event",result=result)
    listToResult(testname,"e_failure",lineFailureDiff,refresultdict,timingdict,toomany,"zero events",result=result)
    listToResult(testname,"f_10pc",tenPercentDiffs,refresultdict,timingdict,toomany,"+/- <10 %",result=result)
    listToResult(testname,"g_Rest",rest,refresultdict,timingdict,toomany,result=result)
    
    for r,v in refdict.iteritems():
        refresultdict[r]=v[1]
    for r,v in thisdict.iteritems():
        timingdict[r]=v[1]
    
    listToResult(testname,"h_Timing",timings,refresultdict,timingdict,toomany,result=result)


def findReferenceBlock(reference, stdout, result, causes,
                       signature_offset=0, signature=None, id=None):
    old = result.annotations.copy()
    BaseTest().findReferenceBlock(
        reference=reference, stdout=stdout, result=result,
        causes=causes, signature_offset=signature_offset)
    added = set(result.annotations) - set(old)
    for key in added:
        if key.endswith('.observed'):
            observed = result[key]
            expected = result[key.replace('.observed', '.expected')]
            diff = list(difflib.ndiff(expected.splitlines(),
                                      observed.splitlines()))
            s = '--- expected\n+++ observed\n' + '\n'.join(diff)
            result[key.replace('.observed', '.diff')] = result.Quote(s)


def _read_options_pkl(filename):
    """Read a .pkl options file produced by gaudirun.py."""
    import cPickle as pickle
    from GaudiKernel.Configurable import (
        DataObjectHandleBase, PrivateToolHandle)

    def sanitize(x):
        """Sanitize non-trivial types."""
        if isinstance(x, (DataObjectHandleBase, PrivateToolHandle)):
            return x.toStringProperty()
        else:
            return x

    with open(filename) as f:
        d = pickle.load(f)
    return {'{} ({})'.format(c.getName(), c.getType()):
            {k: sanitize(v) for k, v in c.getValuedProperties().iteritems()}
            for c in d.values()}


def read_options_pkl(filename):
    """Read a .pkl options file produced by gaudirun.py.

    A new process is started, which reads the options and passes them
    back via shared memory. This is unfortunately necessary as the side
    effects of loading a pickle with options occasionally prevents
    loading another one in the same process.

    """
    import json
    import multiprocessing

    def worker(filename, ret):
        opts = _read_options_pkl(filename)
        ret['data'] = json.dumps(opts)

    with multiprocessing.Manager() as manager:
        ret = manager.dict()
        p = multiprocessing.Process(target=worker, args=(filename, ret))
        p.start()
        p.join()
        try:
            data = ret['data']
        except KeyError:
            raise RuntimeError('Could not open/parse ' + filename)
        return json.loads(data)


def read_options(filename):
    DataObjectHandleBase = str
    PrivateToolHandle = str
    print 'reading', filename
    if filename.endswith('.py'):
        with open(filename) as f:
            return eval(f.read())
    elif filename.endswith('.pkl'):
        with open(filename) as f:
            return read_options_pkl(filename)
    elif filename.endswith('.opts'):
        with open(filename) as f:
            import collections
            matches = re.finditer(r'^([^=]+)\.([^=.]+) = (.+);$', f.read(),
                                  re.MULTILINE)
            opts = collections.defaultdict(dict)
            for m in matches:
                opts[m.group(1)][m.group(2)] = eval(m.group(3))
            return opts
    else:
        raise RuntimeError('{} not supported!'.format(filename))


def diff_options(fn_old, fn_new, exclusions=[]):
    """Return the diff between two Gaudi option files."""

    exclusions = [(re.compile(alg), re.compile(prop))
                  for alg, prop in exclusions]
    def excluded(alg, prop):
        return any(ralg.match(alg) and rprop.match(prop)
                   for ralg, rprop in exclusions)

    def list_props(d):
        import json
        out = []
        for k in sorted(d):
            out.extend(json.dumps(
                {k: d[k]}, indent=4, sort_keys=True).splitlines()[1:-1])
        return out

    def filter_props(alg, props):
        if not exclusions:
            return props
        return {k: v for k, v in props.items() if not excluded(alg, k)}

    def diff(x, y):
        import difflib
        out = []
        for k in sorted(set(x) | set(y)):
            xk = filter_props(k, x[k]) if k in x else []
            yk = filter_props(k, y[k]) if k in y else []
            if xk == yk:
                continue
            out.append('@@ {} @@'.format(k))
            lines = difflib.ndiff(list_props(xk), list_props(yk), None, None)
            out.extend(lines)
            out.append('')
        return out

    d = diff(read_options(fn_old), read_options(fn_new))
    if d:
        out = '--- {}\n+++ {}\n\n{}'.format(fn_old, fn_new, '\n'.join(d))
    else:
        out = ''
    return out


def compareTimingTableEvents2(file1, file2, result, causes, ancestor=None, censor=None, remove=None):
    def process(filename):
        try:
            table = timingtable.read_csv(filename)
        except IOError:
            causes.append('Cannot open file "{}"'.format(filename))
            table = []
        if ancestor:
            table = timingtable.filter_ancestor(table, ancestor)
        if censor:
            table = timingtable.censor_algs(table, censor)
        if remove:
            table = timingtable.censor_algs(table, remove, remove=True)
        lines = timingtable.format_diff_table(table)
        result[filename] = result.Quote(''.join(lines))
        return lines

    lines1 = process(file1)
    lines2 = process(file2)
    if lines1 and lines2:
        diff = timingtable.ndiff(lines1, lines2, file1, file2)
        if diff:
            result['differences'] = result.Quote(diff)
            causes.append('Differences found in timing tables')
