from __future__ import print_function
import os
import re
import difflib
import inspect
from functools import partial
from types import ModuleType
import GaudiKernel.Configurable
import Gaudi.Configuration
from GaudiKernel.ProcessJobOptions import LogFilter, GetConsoleHandler


def indent(lines):
    return len(re.match('^( *)', lines[0]).group(1)) if lines else 0


def reindent(lines, n):
    m = indent(lines)
    return [n * ' ' + line[m:] for line in lines]


def getsourcelines(x, default=None):
    """Return the source code split by lines or default if not possible.
    """
    try:
        return inspect.getsource(x).splitlines()
    except IOError:
        if default is None:
            raise
        return default


def monkeypatch(x, globals_=None):
    def decorator(func):
        old = getattr(x, func.__name__, None)
        source = getsourcelines(func)
        # remove decorator line
        source = source[1:]

        # Some non-essential printout
        print('Patching {}.{} with'.format(x.__name__, func.__name__))
        oldsource = getsourcelines(old, []) if old else []
        if oldsource:
            diffs = difflib.unified_diff(
                oldsource, reindent(source, indent(oldsource)),
                'old', 'new')
            print('\n'.join(diffs), '\n')

        # Create a function in the x.__globals__ scope
        compiled = compile('\n'.join(source), '', 'exec')
        exec_scope = {}
        globs = (globals_ or
                 (old.__globals__ if old else None) or
                 (x.__dict__ if isinstance(x, ModuleType) else {}))
        exec(compiled, globs, exec_scope)
        patched = exec_scope.popitem()[1]
        setattr(x, func.__name__, patched)

        return None
    return decorator


try:
    strictConfUserProp = int(os.environ.get('GAUDI_STRICT_CONF_USER', '0'))
except ValueError:
    strictConfUserProp = 1


OldConfigurableUser = GaudiKernel.Configurable.ConfigurableUser
Configurable = GaudiKernel.Configurable.Configurable


def warn_once(msg):
    log = GaudiKernel.Configurable.log
    try:
        warned = warn_once._warned
    except AttributeError:
        warned = warn_once._warned = set()
    if msg not in warned:
        log.warning(msg)
        log.warning('Downgrading warning to debug')
        warned.add(msg)
    else:
        log.debug(msg)


class PatchedConfigurableUser(OldConfigurableUser):
    __slots__ = {
        '_applying': False,
    }

    def __init__( self, name = Configurable.DefaultName, _enabled = True, **kwargs ):
        super( OldConfigurableUser, self ).__init__( name )
        for n, v in kwargs.items():
            setattr(self, n, v)
        self._enabled = _enabled
        self.__users__ = []
        self._applied = False
        self._applying = False

        # Needed to retrieve the actual class if the declaration in __used_configurables__
        # and  __queried_configurables__ is done with strings.
        from GaudiKernel.ConfigurableDb import getConfigurable as confDbGetConfigurable

        # Set the list of users of the used configurables
        #
        self.__used_instances__ = []
        for used in self.__used_configurables__:
            # By default we want to use the default name of the instances
            # for the used configurables
            used_name = Configurable.DefaultName
            # If the entry in the list is a tuple, we need a named instance
            if type(used) is tuple:
                used, used_name = used # we re-set used to re-use the code below
                if not used_name:
                    used_name = self._instanceName(used)
            # Check is 'used' is a string or not
            if type(used) is str:
                used_class = confDbGetConfigurable(used)
            else:
                used_class = used
            # Instantiate the configurable that we are going to use
            try:
                inst = used_class(name = used_name, _enabled = False)
            except AttributeError:
                # This cover the case where the used configurable is not a
                # ConfigurableUser instance, i.e. id doesn't have the attribute
                # '_enabled'.
                inst = used_class(name = used_name)
            self._ConfigurableUser__addActiveUseOf(inst)
        for queried in self.__queried_configurables__:
            try:
                if type(queried) is str:
                    queried = confDbGetConfigurable(used)
                inst = queried(_enabled = False)
            except AttributeError:
                inst = queried()
            self._ConfigurableUser__addPassiveUseOf(inst)

    def __setattr__(self, name, value):
        if strictConfUserProp > 0 and hasattr(self, '_applied'):
            if self._applied and name in self._properties:
                if hasattr(self, name) and getattr(self, name) == value:
                    warn_once("Re-setting '{}' property of {!r}"
                              .format(name, self))
                else:
                    msg = ("{!r}.{} = {!r} cannot be modified (with {!r}) after "
                           "the configurable has been applied.\n"
                           .format(self, name, self.getProp(name), value))
                    if strictConfUserProp >= 2:
                        raise RuntimeError(msg)
                    else:
                        log.error(msg)
        OldConfigurableUser.__setattr__(self, name, value)

    def getProp(self, name, strict=True):
        """Returns the value of the given property.
        """
        if strictConfUserProp > 0 and strict and GaudiKernel.Configurable._appliedConfigurableUsers_:
            # _appliedConfigurableUsers_ is also set when applying
            if name in self._properties and not self._applied and not self._applying:
                msg = ("{!r}: Configuration cannot be queried before "
                       "the configurable has been applied.".format(self))
                if strictConfUserProp >= 2:
                    raise RuntimeError(msg)
                else:
                    log.error(msg)
        return OldConfigurableUser.getProp(self, name)


print('Patching ConfigurableUser')
GaudiKernel.Configurable.ConfigurableUser = PatchedConfigurableUser
Gaudi.Configuration.ConfigurableUser = PatchedConfigurableUser


@monkeypatch(GaudiKernel.Configurable)
def applyConfigurableUsers():
    """
    Call the apply method of all the ConfigurableUser instances respecting the
    dependencies. First the C.U.s that are not used by anybody, then the used
    ones, when they are not used anymore.
    """
    # Avoid double calls
    global _appliedConfigurableUsers_, postConfigActions
    if _appliedConfigurableUsers_:
        return
    _appliedConfigurableUsers_ = True

    def applicableConfUsers():
        '''
        Generator returning all the configurables that can be applied in the
        order in which they can be applied.
        '''
        # This is tricky...
        # We keep on looking for the first configurable that can be applied.
        # When we cannot find any, 'next()' raises a StopIteration that is
        # propagated outside of the infinite loop and the function, then handled
        # correctly from outside (it is the exception that is raised when you
        # exit from a generator).
        # Checking every time the full list is inefficient, but it is the
        # easiest way to fix bug #103803.
        # <https://savannah.cern.ch/bugs/?103803>
        while True:
            yield (c for c in Configurable.allConfigurables.values()
                   if c.isApplicable()).next()

    debugApplyOrder = 'GAUDI_DUBUG_CONF_USER' in os.environ
    for c in applicableConfUsers():
        if c._enabled:
            log.info("applying configuration of %s", c.name())
            if debugApplyOrder:
                sys.stderr.write('applying %r' % c)
            c._applying = True
            c.__apply_configuration__()
            c._applying = False
            log.info(c)
        else:
            log.info("skipping configuration of %s", c.name())
        c._applied = True # flag the instance as already applied
        if hasattr(c, "__detach_used__"):
            # tells the used configurables that they are not needed anymore
            c.__detach_used__()

    # check for dependency loops
    leftConfUsers = [c for c in Configurable.allConfigurables.values()
                     if hasattr(c, '__apply_configuration__') and
                        c._enabled and not c._applied]
    # if an enabled configurable has not been applied, there must be a dependency loop
    if leftConfUsers:
        raise Error("Detected loop in the ConfigurableUser"
                    " dependencies: %r" % [ c.name()
                                            for c in leftConfUsers ])
    # ensure that all the Handles have been triggered
    known = set()
    unknown = set(Configurable.allConfigurables)
    while unknown:
        for k in unknown:
            if not known: # do not print during the first iteration
                log.debug('new configurable created automatically: %s', k)
            # this trigger the instantiation from handles
            Configurable.allConfigurables[k].properties()
            known.add(k)
        unknown -= known
    # Call post-config actions
    for action in postConfigActions:
        action()


@monkeypatch(GaudiKernel.Configurable)
def applyConfigurableUsers_old():
    """
    Obsolete (buggy) implementation of applyConfigurableUsers(), kept to provide
    backward compatibility for configurations that where relying (implicitly) on
    bug #103803, or on a specific (non guaranteed) order of execution.

    @see applyConfigurableUsers()
    """
    # Avoid double calls
    global _appliedConfigurableUsers_, postConfigActions
    if _appliedConfigurableUsers_:
        return
    _appliedConfigurableUsers_ = True

    debugApplyOrder = 'GAUDI_DUBUG_CONF_USER' in os.environ
    confUsers = [ c
                  for c in Configurable.allConfigurables.values()
                  if hasattr(c,"__apply_configuration__") ]
    applied = True # needed to detect dependency loops
    while applied and confUsers:
        newConfUsers = [] # list of conf users that cannot be applied yet
        applied = False
        for c in confUsers:
            if hasattr(c,"__users__") and c.__users__:
                newConfUsers.append(c) # cannot use this one yet
            else: # it does not have users or the list is empty
                applied = True
                # the ConfigurableUser is enabled if it doesn't have an _enabled
                # property or its value is True
                enabled = (not hasattr(c, "_enabled")) or c._enabled
                if enabled:
                    log.info("applying configuration of %s", c.name())
                    if debugApplyOrder:
                        sys.stderr.write('applying %r' % c)
                    c._applying = True
                    c.__apply_configuration__()
                    c._applying = False
                    log.info(c)
                else:
                    log.info("skipping configuration of %s", c.name())
                c._applied = True  # flag the instance as already applied
                if hasattr(c, "__detach_used__"):
                    # tells the used configurables that they are not needed anymore
                    c.__detach_used__()
        confUsers = newConfUsers # list of C.U.s still to go
    if confUsers:
        # this means that some C.U.s could not be applied because of a dependency loop
        raise Error("Detected loop in the ConfigurableUser "
                    " dependencies: %r" % [ c.name()
                                            for c in confUsers ])
    # ensure that all the Handles have been triggered
    known = set()
    unknown = set(Configurable.allConfigurables)
    while unknown:
        for k in unknown:
            if not known: # do not print during the first iteration
                log.debug('new configurable created automatically: %s', k)
            # this trigger the instantiation from handles
            Configurable.allConfigurables[k].properties()
            known.add(k)
        unknown -= known
    # Call post-config actions
    for action in postConfigActions:
        action()


def printOn(self, step = 1, force = False):
    """
    Decrease the printing_level of 'step' units. ( >0 means no print)
    The level cannot go below 0, unless the force flag is set to True.
    A negative value of the threshold disables subsequent "PrintOff"s.
    """
    if force:
        self.printing_level -= step
    else:
        if self.printing_level >= step:  # RM
            self.printing_level -= step
        else:
            raise ValueError('Do not call printOn without force=True!')  # RM
            self.printing_level = 0

LogFilter.printOn = printOn
GetConsoleHandler()._filter.printOn = partial(printOn, GetConsoleHandler()._filter)
