#!/usr/bin/env python
"""Script to test the rates of a HLT configuration."""

import os
import re
import atexit
import argparse
import pickle

import Gaudi.Configuration
from Configurables import GaudiSequencer
from Gaudi.Configuration import appendPostConfigAction
from Configurables import ApplicationMgr, NTupleSvc
from Configurables import LHCbApp
from GaudiPython import AppMgr
from DecisionTuple import LumiTuple, L0Tuple, Hlt1Tuple, Hlt2Tuple, StrippingTuple


__author__ = (
    'Mika Vesterinen',
    'Rosen Matev',
    'Sascha Stahl'
)


def get_properties(algname):
    return {k: v.value() for k, v in
            AppMgr().algorithm(algname).properties().iteritems()}


def is_turbopp_line(line):
    code = get_properties('Hlt2PersistRecoLineFilter')['Code']
    return (line + 'Decision') in code


def is_turbo_line(line):
    return get_properties(line)['Turbo']


def get_lines(stage):
    return [i.split('/')[-1] for i in get_properties(stage)['Members']]


parser = argparse.ArgumentParser(description="Run the rate test")
parser.add_argument("name", help="Name (prefix) for all test products")
# general Moore steering
parser.add_argument("-n", "--evtmax", type=int, default=50000,
                    help="Number of events to run over")
parser.add_argument("-d", "--datatype", dest="DataType", default="2017",
                    help="DataType for LHCbApp",)
# options related to the input data
parser.add_argument("--inputdata", default="2017NB_L0Filt0x1707",
                    help="TestFileDB label, path to raw file or path to options file.")
parser.add_argument("--catalog", default="",
                    help="Path to xml catalog file.")
parser.add_argument("--TFDBForTags",
                    help="Original test file db key, to pick up tags")
parser.add_argument("--dddbtag",
                    help="DDDBTag to use. Rely on TestFileDB if unspecified.")
parser.add_argument("--conddbtag",
                    help="CondDBTag to use. Rely on TestFileDB if unspecified.")
parser.add_argument("--Simulation", action="store_true",
                    help="Is this simulated data?")
# other options
parser.add_argument("--hlt1", action="store_true", help="Run on HLT1 output")
parser.add_argument("--input_rate", type=float, default=1000,
                    help="Input rate from L0 in Hz")
parser.add_argument("--slines", help="File with stripping lines")

parser.add_argument("--OutputLevel", type=int, default=4, help="Output level")
args = parser.parse_args()


theApp = LHCbApp()
theApp.EvtMax = args.evtmax
theApp.DataType = args.DataType
#theApp.OutputLevel = args.OutputLevel

if args.slines:
    with open(args.slines, "rb") as f:
        stripping_lines = pickle.load(f)

# inputdata
try:
    from PRConfig import TestFileDB
    TestFileDB.test_file_db[args.inputdata].run(configurable=thaApp)
except KeyError:
    from GaudiConf import IOHelper
    ext = os.path.splitext(args.inputdata)[1].lower()
    if ext == ".py":
        from LHCbKernel.Configuration import importOptions
        importOptions(args.inputdata)
        if args.catalog:
            from Gaudi.Configuration import FileCatalog
            FileCatalog().Catalogs = ['xmlcatalog_file:' + args.catalog]
    elif ext in ['.mdf', '.raw']:
        IOHelper("MDF").inputFiles(["mdf:" + args.inputdata])
    else:
        IOHelper().inputFiles(["PFN:" + args.inputdata])

# Only set the tags if specified
# This should be done if you neither provide TestFileDB label via --inputdata, nor provide it via --TFDBForTags
if args.TFDBForTags:
    from PRConfig import TestFileDB
    TestFileDB.test_file_db[args.TFDBForTags].setqualifiers(configurable=Moore())
if args.conddbtag:
    theApp.CondDBtag = args.conddbtag
if args.dddbtag:
    theApp.DDDBtag = args.dddbtag


# Emulate minumum possible configuration from DaVinci() and friends
from Configurables import ApplicationMgr
# - configure data on demand (e.g. for ODIN/HltLumiSummary bank decoding)
from Configurables import DecodeRawEvent, DataOnDemandSvc, L0Conf
DecodeRawEvent().DataOnDemand = True
ApplicationMgr().ExtSvc += [DataOnDemandSvc()]
L0Conf().EnableL0DecodingOnDemand = True
# - configure timing table
from Configurables import ToolSvc, AuditorSvc, TimingAuditor, SequencerTimerTool
ApplicationMgr().AuditAlgorithms = True
AuditorSvc().Auditors += ['TimingAuditor']
TimingAuditor().addTool(SequencerTimerTool, name="TIMER")
TimingAuditor().TIMER.NameSize = 50
ApplicationMgr().ExtSvc += [ToolSvc(), AuditorSvc()]
from Configurables import EventSelector
EventSelector().PrintFreq = 1000
# - configure output
from Configurables import NTupleSvc
ApplicationMgr().HistogramPersistency = "ROOT"
ApplicationMgr().ExtSvc +=  [ NTupleSvc() ]
NTupleSvc().Output = ["FILE1 DATAFILE='{}' TYP='ROOT' OPT='NEW'"
                      .format(args.name + '-tuples.root')]


from Configurables import HltRoutingBitsFilter, FilterOnLumiSummary
content_seq = GaudiSequencer("ContentSeq", Members=[
    'L0Tuple', 'Hlt1Tuple'])
if not args.hlt1:
    content_filter = HltRoutingBitsFilter("ContentFilter")
    content_filter.RequireMask = [0x0, 0x0, 0x80000000]
    content_seq.Members.insert(0, content_filter)
    content_seq.Members.append('Hlt2Tuple')
# TODO lumi only if on raw
lumi_filter = FilterOnLumiSummary('LumiFilter', CounterName="Random",
                                  ValueName="RandomMethod")
lumi_seq = GaudiSequencer("LumiSeq", Members=[lumi_filter, 'LumiTuple'])
ApplicationMgr().TopAlg = [content_seq, lumi_seq]

# Instantiate the AppMgr such that we can instantiate algorithms
gaudi = AppMgr()

# Instantiate GuadiPython algorithms before initialize if you're trying
# to insert them in some sequence!

# this raw location is a heuristic...
raw_location_kwarg = {}
if os.path.splitext(args.inputdata)[1].lower() not in ['.mdf', '.raw']:
    raw_location_kwarg['raw_location_kwarg'] = 'Trigger/RawEvent'
lumi = LumiTuple('LumiTuple', 'Lumi')
l0 = L0Tuple('L0Tuple', 'L0', **raw_location_kwarg)
hlt1 = Hlt1Tuple('Hlt1Tuple', 'Hlt1', **raw_location_kwarg)
hlt2 = Hlt2Tuple('Hlt2Tuple', 'Hlt2', **raw_location_kwarg)
if args.slines:
    stripping = StrippingTuple('StrippingTuple', 'Strip', stripping_lines)

gaudi.initialize()

# Append algorithms to "TopAlg"
# gaudi.addAlgorithm(l0)
# gaudi.addAlgorithm(hlt1)
# gaudi.addAlgorithm(hlt2)
if args.slines:
    gaudi.addAlgorithm(stripping)

# Run
atexit.register(gaudi.exit)
gaudi.run(args.evtmax)
gaudi.stop()
gaudi.finalize()
