import re
import GaudiPython
from GaudiPython.GaudiAlgs import TupleAlgo
LHCb = GaudiPython.gbl.LHCb

__author__ = 'Rosen Matev'

__all__ = (
    'L0Tuple',
    'Hlt1Tuple',
    'Hlt2Tuple',
)

DECISION_SUFFIX = re.compile(r'Decision$')

RAW_BANK_TYPES = [(i, LHCb.RawBank.typeName(i))
                  for i in range(LHCb.RawBank.LastType)]


def rawbank_sizes(rawevent):
    """Return (name, size) for each raw bank type."""
    if rawevent:
        def size(i):
            return sum(bank.totalSize() for bank in rawevent.banks(i))
    else:
        def size(i):
            return 0
    return [(name, size(i)) for i, name in RAW_BANK_TYPES]


def routing_bits(rawevent):
    """Return a list with the 96 routing bit values."""
    if not rawevent:
        return [0]*96
    rbbanks = rawevent.banks(LHCb.RawBank.HltRoutingBits)
    if rbbanks.size() < 1:
        return [0]*96  # some events don't have rb bank?
    assert rbbanks.size() == 1, "Multiple routing bit banks!"
    d = rbbanks[rbbanks.size() - 1].data()  # get last bank
    bits = "{:032b}{:032b}{:032b}".format(d[2], d[1], d[0])
    return map(int, reversed(bits))


def fill_odin(tup, odin):
    tup.column('run', odin.runNumber())
    tup.column_ull('gps', odin.gpsTime())
    tup.column_ull('evt', odin.eventNumber())
    tup.column('bcid', odin.bunchId())
    tup.column('bx', odin.bunchCrossingType())


def fill_bits(tup, rawevent, i0, i1, prefix='rb_'):
    bits = routing_bits(rawevent)
    for i in xrange(i0, i1):
        tup.column(prefix + str(i), bits[i])


def fill_sizes(tup, rawevent, prefix='rawevent_'):
    sizes = rawbank_sizes(rawevent)
    tup.column(prefix + 'total', sum(s[1] for s in sizes))
    for name, size in sizes:
        tup.column(prefix + name, size)


L0_BX_MAP = [
    (0, ''), (-2, '_Prev2'), (-1, '_Prev1'),
    (1, '_Next1'), (2, '_Next2'),
]
L0_DECISIONS_MAP = [(1, 'Physics'), (2, 'Beam1'), (4, 'Beam2')]


def fill_l0du(tup, l0du):
    tup.column('TCK', l0du.tck())
    for dec, decname in L0_DECISIONS_MAP:
        for bx, bxname in L0_BX_MAP:
            tup.column('L0{}{}'.format(decname, bxname),
                       l0du.decisionFromSummary(dec, bx))
    assert l0du.valid()
    l0du_conf = l0du.configuration()
    channels = dict(l0du_conf.channels())
    for ch in sorted(channels.keys()):
        for bx, bxname in L0_BX_MAP:
            tup.column('L0{}{}'.format(ch, bxname),
                       l0du.channelDecisionByName(ch, bx))


def fill_hdr(tup, hdr):
    tup.column('TCK', hdr.configuredTCK())
    tup.column('TaskID', hdr.taskID())
    for name, dec in sorted(map(tuple, hdr.decReports())):
        tup.column(DECISION_SUFFIX.sub('', name), dec.decision())
        tup.column(DECISION_SUFFIX.sub('Candidates', name),
                   dec.numberOfCandidates())


class DecisionTuple(TupleAlgo):
    def __init__(self, name, tuple_name, raw_location="DAQ/RawEvent"):
        TupleAlgo.__init__(self, name, NTupleLUN='FILE1', NTupleDir='')
        self._tuple_name = tuple_name
        self._raw_location = raw_location

    def initialize(self):
        status = TupleAlgo.initialize(self)
        if status.isFailure():
            return status
        self._tuple = self.nTuple(self._tuple_name)
        return GaudiPython.SUCCESS

    def finalise(self):
        del self.tuple
        return TupleAlgo.finalise(self)

    def get_strict(self, location):
        x = self.get(location)
        if not x:
            raise RuntimeError('Object not found in TES at "{}"'
                               .format(location))
        return x

    def execute(self):
        write = self.fill(self._tuple)
        if write:
            self._tuple.write()
        return GaudiPython.SUCCESS


class LumiTuple(DecisionTuple):
    def fill(self, tup):
        fill_odin(tup, self.get_strict('DAQ/ODIN'))
        rawevent = self.get_strict(self._raw_location)
        fill_bits(tup, rawevent, 0, 96)
        fill_sizes(tup, rawevent)
        return True


class L0Tuple(DecisionTuple):
    def fill(self, tup):
        l0du = self.get('Trig/L0/L0DUReport')
        if not l0du:
            return False
        fill_odin(tup, self.get_strict('DAQ/ODIN'))
        fill_l0du(tup, l0du)
        return True


class Hlt1Tuple(DecisionTuple):
    def fill(self, tup):
        hdr = self.get('Hlt1/DecReports')
        if not hdr or not hdr.decReport('Hlt1Global').decision():
            return False
        fill_odin(tup, self.get_strict('DAQ/ODIN'))
        fill_hdr(tup, hdr)
        rawevent = self.get_strict(self._raw_location)
        fill_bits(tup, rawevent, 0, 64)
        fill_sizes(tup, rawevent)
        return True


class Hlt2Tuple(DecisionTuple):
    def __init__(self, name, tuple_name, streams=[''], raw_location="DAQ/RawEvent"):
        DecisionTuple.__init__(self, name, tuple_name, raw_location)
        self.streams = streams

    def fill(self, tup):
        hdr = self.get('Hlt2/DecReports')
        if not hdr or not hdr.decReport('Hlt2Global').decision():
            return False
        fill_odin(tup, self.get_strict('DAQ/ODIN'))
        fill_hdr(tup, hdr)
        for stream in self.streams:
            rawevent = self.get(
                self._raw_location.replace('RawEvent', stream + 'RawEvent'))
            s = (stream + '_') if stream else ''
            fill_bits(tup, rawevent, 64, 96, prefix='rb_' + s)
            fill_sizes(tup, rawevent, prefix='rawevent_' + s)
        return True

class StrippingTuple(DecisionTuple):
    def __init__(self, name, tuple_name, lines):
        DecisionTuple.__init__(self, name, tuple_name)
        self.lines = sorted(lines)

    def fill(self, tup):
        fill_odin(tup, self.get_strict('DAQ/ODIN'))
        sdr = self.get('Strip/Phys/DecReports')
        self.fill_sdr(tup, sdr)
        return True

    def fill_sdr(self, tup, sdr):
        for line in self.lines:
            tup.column(line, 0)
        for name, dec in sorted(map(tuple, sdr.decReports())):
            if "StrippingStream" not in name:
                tup.column(DECISION_SUFFIX.sub('', name), 1)
