#!/usr/bin/env python
"""Script to test the rates of a HLT configuration."""

import os
import re
import atexit
import argparse

import Gaudi.Configuration
from Configurables import GaudiSequencer
from Configurables import EventSelector, HltConf
from Moore.Configuration import Moore
from Gaudi.Configuration import appendPostConfigAction
from GaudiKernel.Configurable import applyConfigurableUsers_old
from Configurables import ApplicationMgr, NTupleSvc

from GaudiPython import AppMgr
from DecisionTuple import L0Tuple, Hlt1Tuple, Hlt2Tuple


__author__ = (
    'Mika Vesterinen',
    'Rosen Matev',
)


def get_properties(algname):
    return {k: v.value() for k, v in
            AppMgr().algorithm(algname).properties().iteritems()}


def is_turbopp_line(line):
    code = get_properties('Hlt2PersistRecoLineFilter')['Code']
    return (line + 'Decision') in code


def is_turbo_line(line):
    return get_properties(line)['Turbo']


def get_lines(stage):
    return [i.split('/')[-1] for i in get_properties(stage).get('Members', [])]


parser = argparse.ArgumentParser(description="Run the rate test")
parser.add_argument("name", help="Name (prefix) for all test products")
# general Moore steering
parser.add_argument("-n", "--evtmax", type=int, default=30000,
                    help="Number of events to run over")
parser.add_argument("--settings",
                    help="ThresholdSettings used by Moore")
parser.add_argument("--TCK", help="Run from TCK instead of settings")
parser.add_argument("--split", default="", choices=["", "Hlt1", "Hlt2"],
                    help="Split mode")
parser.add_argument("-d", "--datatype", dest="DataType",
                    help="DataType for Moore")
# options related to the input data
parser.add_argument("--inputdata", default="2017NB_L0Filt0x1707",
                    help="TestFileDB label, or path to raw file.")
parser.add_argument("--TFDBForTags",
                    help="Original test file db key, to pick up tags")
parser.add_argument("--dddbtag",
                    help="DDDBTag to use. Rely on TestFileDB if unspecified.")
parser.add_argument("--conddbtag",
                    help="CondDBTag to use. Rely on TestFileDB if unspecified.")
parser.add_argument("--Simulation", action="store_true",
                    help="Is this simulated data?")
parser.add_argument("--Online", action="store_true",
                    help="Are you running on plus on data from the farm?")
parser.add_argument("--L0TCK",
                    help=("L0 TCK. Should specificy if the input data was "
                          "filtered with something different to what is "
                          "in the ThresholdSettings."))
# other options
parser.add_argument("--TCKData", help="Directory containing config.cdb")
parser.add_argument("--DisableRunStampCheck", action="store_true",
                    help="DisableRunStampCheck")
parser.add_argument("--DisableOutputStreaming", action="store_true",
                    help="EnableOutputStreaming")
parser.add_argument("--OutputLevel", type=int, default=4, help="Output level")
parser.add_argument("--AdditionalHlt2Lines", default="",
                    help="Semicolon separated list of lines")
args = parser.parse_args()

if bool(args.TCK) == bool(args.settings):
    parser.error("You must specify either --TCK or --settings")


# By default, we remove the following lines
# - In the typical input dataset, we won't get the correct rate of lumi events
# - In the NB datasets, Hlt1MBNoBias will always fire (with its 10% prescale)
# - Also avoid the rate limited lines
EXCLUDED_LINES = [
    "Hlt1MBNoBias",
    "Hlt1MBNoBiasRateLimited",
    "Hlt1Lumi",
    "Hlt1CharmXSec",
]

if args.TCK:
    # automaticaly determine split mode
    level = int(args.TCK, 16) >> 28 & 3
    # TODO warn if they don't match?
    args.split = {1: 'Hlt1', 2: 'Hlt2'}[level]

    Moore().UseTCK = True
    Moore().InitialTCK = args.TCK
    if args.TCKData:
        Moore().TCKData = args.TCKData
    if args.split == "Hlt1":
        from Configurables import HltConfigSvc
        trans = {'DeterministicPrescaler/{}PreScaler'.format(n):
                 {'AcceptFraction': {'.*': '0'}}
                 for n in EXCLUDED_LINES}
        HltConfigSvc().ApplyTransformation = trans
else:
    from Configurables import HltConf

    Moore().UseTCK = False
    if args.settings:
        Moore().ThresholdSettings = args.settings

    # if specified, we can overide the L0TCK in the ThresholdSettings
    # to use the one with which the input data set was filtered
    if args.L0TCK:
        HltConf().L0TCK = args.L0TCK

    # by default, the output streaming is disabled in HltConf
    Moore().EnableOutputStreaming = not args.DisableOutputStreaming

    HltConf().RemoveHlt1Lines = EXCLUDED_LINES

    # option to add Hlt2 lines that aren't in the settings
    # e.g. if you are developing a new line
    if args.AdditionalHlt2Lines:
        HltConf().AdditionalHlt2Lines = args.AdditionalHlt2Lines.split(";")


# Settings that are common to "TCK" and "ThresholdSettings" modes
if args.DataType:
    Moore().DataType = args.DataType
Moore().Split = args.split
Moore().RemoveInputHltRawBanks = Moore().Split in ["", "Hlt1"]
Moore().CheckOdin = False
Moore().ForceSingleL0Configuration = False
Moore().EnableTimer = True
Moore().EvtMax = args.evtmax
Moore().OutputLevel = args.OutputLevel
Moore().outputFile = args.name + '.mdf'
Moore().TimerSummaryFile = args.name + '.csv'
if not args.TCK:
    from Configurables import HltMonitoringConf
    HltMonitoringConf().OutputFile = args.name + "-histos.root"
else:
    pass
    # TODO figure out how enable histo saving, if at all possible

# Needed when running over real data
if not args.Simulation:
    from Configurables import CondDB
    CondDB().IgnoreHeartBeat = True
    CondDB().EnableRunChangeHandler = True
    CondDB().EnableRunStampCheck = not args.DisableRunStampCheck
    if args.Online:
        CondDB().UseDBSnapshot = True
        CondDB().DBSnapshotDirectory = "/group/online/hlt/conditions"
        CondDB().Tags["ONLINE"] = 'fake'
        CondDB().Online = True
        CondDB().EnableRunStampCheck = False

        import sys
        rd = '/group/online/hlt/conditions/RunChangeHandler'
        sys.path.append(rd)
        if args.split == "Hlt2":
            import All
            CondDB().RunChangeHandlerConditions = All.ConditionMap
        if args.split == "Hlt1":
            import AllHlt1
            CondDB().RunChangeHandlerConditions = AllHlt1.ConditionMap

# inputdata
try:
    from PRConfig import TestFileDB
    TestFileDB.test_file_db[args.inputdata].run(configurable=Moore())
except KeyError:
    from GaudiConf import IOHelper
    if ".mdf" in args.inputdata:
        IOHelper("MDF").inputFiles([args.inputdata])
    elif ".dst" in args.inputdata:
        IOHelper().inputFiles(["PFN:" + args.inputdata])
    Moore().Simulation = args.Simulation

# Only set the tags if specified
# This should be done if you neither provide TestFileDB label via --inputdata, nor provide it via --TFDBForTags
if args.TFDBForTags:
    from PRConfig import TestFileDB
    TestFileDB.test_file_db[args.TFDBForTags].setqualifiers(configurable=Moore())
if args.conddbtag:
    Moore().CondDBtag = args.conddbtag
if args.dddbtag:
    Moore().DDDBtag = args.dddbtag

# deal with the Herschel decoding issue on MC (only when running from settings)
if Moore().getProp('Simulation') and not args.TCK:
    from Configurables import HltConf
    remove=["Hlt1LowMultMaxVeloAndHerschel", "Hlt1LowMultVeloAndHerschel_Leptons", "Hlt1LowMultVeloAndHerschel_Hadrons","Hlt1LowMultMaxVeloAndHerschel","Hlt1LowMultVeloAndHerschel_Hadrons", "Hlt1LowMultVeloAndHerschel_Leptons"]
    print '#'*100
    print 'WARNING: SINCE YOU RUN ON A MC SAMPLE, WE ASSUME THAT WE NEED TO DISABLE ALL HLT1 LINES THAT REQUIRE DECODING OF HERSCHEL BANKS'
    print 'DISABLING:', remove
    print '#'*100
    HltConf().RemoveHlt1Lines += remove


# print the Moore configurations
print args
print Moore()


# NOTE: GaudiPython calls applyConfigurableUsers() by default and
# we want to be consistent with gaudirun!
applyConfigurableUsers_old()

# When we run in unsplit mode from settings, insert the Hlt1 tuple
# algorithm before the Hlt2 sequence
if args.split == '':
    assert not args.TCK
    seq = GaudiSequencer('HltDecisionSequence')
    i = seq.Members.index(GaudiSequencer('Hlt2'))
    seq.Members.insert(i, 'Hlt1Tuple')

# Setup ntuple output
ApplicationMgr().ExtSvc += [NTupleSvc()]
NTupleSvc().Output = ["FILE1 DATAFILE='{}' TYP='ROOT' OPT='NEW'"
                      .format(args.name + '-tuples.root')]

# Instantiate the AppMgr such that we can instantiate algorithms
gaudi = AppMgr()

# Instantiate GuadiPython algorithms before initialize if you're trying
# to insert them in some sequence!
l0 = L0Tuple('L0Tuple', 'L0')
if args.split in ['', 'Hlt1']:
    hlt1 = Hlt1Tuple('Hlt1Tuple', 'Hlt1')
if args.split in ['', 'Hlt2']:
    hlt2 = Hlt2Tuple('Hlt2Tuple', 'Hlt2')

gaudi.initialize()

# After initialize, we can query the confgiuration, which works when
# running from settings and from TCK
lines = {
    "Hlt1": set(get_lines('Hlt1')),
    "Hlt2": set(get_lines('Hlt2')),
}
persistRecoLines = [line for line in lines['Hlt2']
                    if is_turbopp_line(line)]
streams = get_properties('HltOutputSvc')['StreamGroups'].keys()

# Append algorithms to "TopAlg"
gaudi.addAlgorithm(l0)
if args.split == 'Hlt1':  # '' is handled above
    gaudi.addAlgorithm(hlt1)
if args.split in ['', 'Hlt2']:
    hlt2.streams = streams
    gaudi.addAlgorithm(hlt2)

# Run
atexit.register(gaudi.exit)
gaudi.run(args.evtmax)
gaudi.stop()
gaudi.finalize()
