"""
High-level configuration tool for running L0Emulation
"""


from Gaudi.Configuration import *
from Configurables import GaudiSequencer
from LHCbKernel.Configuration import *
from Configurables import LHCbApp, L0Conf, DecodeRawEvent
from Configurables import ApplicationMgr
from GaudiConf import IOExtension, IOHelper

class L0App(LHCbConfigurableUser):
    ## Possible used Configurables
    __used_configurables__ = [ LHCbApp
                             , L0Conf
                             , DecodeRawEvent ]
    
    __slots__ = {
          "EvtMax":            -1    # Maximum number of events to process
        , "SkipEvents":        0
        , "Simulation":        True # True implies use SimCond
        , "DataType":          '2018' # Data type, can be [ 'DC06','2008' ]
        , "DDDBtag" :          'default' # default as set in DDDBConf for DataType
        , "CondDBtag" :        'default' # default as set in DDDBConf for DataType
        , "ReplaceL0Banks" : False # remove existing banks first?
        , "ReplayL0DUOnly": False #
        , 'TCK' : '' #"Can be a L0TCK or a full TCK, it doesn't matter"
        , "outputFile" :       '' # output filename
        , 'WriteFSR'    :  True #copy FSRs as required
        , 'MergeGenFSR' : False
        , 'EventPreFilters': []
        , 'Monitor': True
        , 'L0Filter': ''
        , 'EmulateHC': False
        }
    _propertyDocDct ={
          "EvtMax":  "Maximum number of events to process, default all"
        , "SkipEvents":   "Events to skip, default     0"
        , "Simulation":  "True implies use SimCond"
        , "DataType":   "Data type, can be [ 'DC06','2008' ...]"
        , "DDDBtag" :  "Databse tag, default as set in DDDBConf for DataType"
        , "CondDBtag" : "Databse tag, default as set in DDDBConf for DataType"
        , "ReplaceL0Banks" : "Whether to replace the existing L0 banks"
        , "ReplayL0DUOnly": "Use existing L0Calo and L0Muon data and recompute L0DU decision"
        , 'TCK' : "Can be a L0TCK or a full TCK, it doesn't matter"
        , "outputFile" :       'output filename, automatically selects MDF or InputCopyStream'
        , 'WriteFSR'    :  'copy FSRs as required'
        , 'MergeGenFSR' : 'Merge the generator FSRs'
        , 'EventPreFilters': 'List of filters to be executed before the L0 sequence'
        , 'Monitor': 'Whether to run the L0DUReportMonitor'
        , 'L0Filter': 'LoKi L0Filter code to be applied before writing'
        }
    def _safeSet(self,conf,param):
        """
        Propagate a property only if it is set, and only if the underlying configurable has not been set
        """
        for p in param:
            if (not self.isPropertySet(p)) or conf.isPropertySet(p):
                #print "skipping", p
                continue
            
            conf.setProp(p,self.getProp(p))
            #print "setting", p
    
    def _setRawEventLocations(self):
        """
        Copied from Moore, check the raw event locations are set correctly
        """
        if (not DecodeRawEvent().isPropertySet("OverrideInputs")) or DecodeRawEvent().getProp("OverrideInputs") is None:
            #default version which comes out of the Pit,
            #currently just DAQ/RawEvent
            DecodeRawEvent().OverrideInputs="Pit" 

    def _setupLoKi(self):
        from Configurables import LoKiSvc
        LoKiSvc().Welcome = False

    def _configureOutput(self, requirements=[]):
        """ Copied from Moore, handle output to DST or MDF
        """
        fname = self.getProp('outputFile')
        if not fname:
            return

        writerSeq = GaudiSequencer('WriterSeq')

        if requirements:
            from Configurables import LoKi__VoidFilter as VoidFilter
            self._setupLoKi()
            code = ' & '.join(["ALG_EXECUTED('{0}') & ALG_PASSED('{0}')".format(i)
                               for i in requirements])
            filterAlg = VoidFilter('WriterFilter', Code=code, Preambulo=[
                'from LoKiHlt.algorithms import ALG_EXECUTED, ALG_PASSED'])
            writerSeq.Members.append(filterAlg)

        #check the file type and use MDF writer or InputCopyStream
        outputAlgs = []
        if IOExtension().detectFileType(fname) == 'MDF':
            from Configurables import LHCb__MDFWriter as MDFWriter
            writer = MDFWriter( 'Writer'
                              , Compress = 0
                              , ChecksumType = 1
                              , GenerateMD5 = True
                              , Connection = 'file://' + fname
                              )
            outputAlgs.extend(IOHelper("MDF", "MDF").outputAlgs(
                fname, writer, writeFSR=False))
        else : 
            if self.getProp('WriteFSR') and self.getProp('MergeGenFSR'):
                from Configurables import GenFSRMerge
                outputAlgs.append(GenFSRMerge())

            from Configurables import InputCopyStream
            writer = InputCopyStream("Writer")
            outputAlgs.extend(IOHelper().outputAlgs(
                fname, writer, writeFSR=self.getProp('WriteFSR')))

        writerSeq.Members += outputAlgs
        ApplicationMgr().OutStream += [writerSeq]

    def __apply_configuration__(self):
        self._safeSet( LHCbApp(), ['EvtMax','SkipEvents','Simulation', 'DataType' , 'CondDBtag','DDDBtag'] )

        ApplicationMgr().AppName="L0App within Moore"

        from Configurables import DataOnDemandSvc
        dod = DataOnDemandSvc()
        if dod not in ApplicationMgr().ExtSvc :
            ApplicationMgr().ExtSvc.append( dod )
        
        DecodeRawEvent().DataOnDemand=True

        importOptions( "$L0TCK/L0DUConfig.opts" )
        
        self._setRawEventLocations()
        
        from DAQSys.Decoders import DecoderDB
        l0du=DecoderDB["L0DUFromRawTool"]
        l0du.Properties["StatusOnTES"] = False

        mainSeq = GaudiSequencer('MainSeq')
        if self.getProp('EventPreFilters'):
            prefilterSeq = GaudiSequencer(
                'EventPreFilterSeq', Members=self.getProp('EventPreFilters'))
            mainSeq.Members.append(prefilterSeq)

        #configure L0 Sequence
        l0seq = GaudiSequencer("L0")
        L0TCK = '0x%s' % self.getProp('TCK')[-4:]
        if ( self.getProp("ReplayL0DUOnly") ) and ( self.getProp("ReplaceL0Banks") ):
            # Decode the l0du to produce the processor data (input to simulation)
            from L0DU.L0Algs import decodeL0DU
            decoding = decodeL0DU()
            decoding.WriteProcData = True
            decoding.WriteOnTES = False # Do not write the L0DU report
            decoding.ProcessorDataLocation  = "Trig/L0/L0DUData"
            # Remove L0DU bank
            from Configurables import bankKiller
            removebanks=bankKiller( "RemoveL0Banks" )
            removebanks.BankTypes = ["L0DU"]
            # Emulate the l0du from the processor data
            from L0DU.L0Algs import emulateL0DU
            emulation = emulateL0DU()
            emulation.ProcessorDataLocations = ["Trig/L0/L0DUData"]
            emulation.WriteBanks = True #
            emulation.WriteOnTES = True # Write the L0DU report
            emulation.TCK = L0TCK
            l0seq.Members += [decoding, removebanks, emulation]

            if self.getProp('EmulateHC'):
                # Manually applying L0Conf is very bad, but should do...
                self.setOtherProps(L0Conf(), ["DataType"])
                L0Conf().__apply_configuration__()
                l0seq.Members.insert(0, L0Conf().l0hcEmulationSeq())
                # The output of hc_emulation is hardcoded
                LHCb__L0ProcessorDataLocation__HC = "Trig/L0/L0DUHCData"
                emulation.ProcessorDataLocations.insert(0, LHCb__L0ProcessorDataLocation__HC)
        else:
            L0Conf().setProp( "TCK", L0TCK )
            L0Conf().setProp( "L0Sequencer", l0seq )
            self.setOtherProps( L0Conf(), [ "DataType" ] )
            if ( self.getProp("ReplaceL0Banks") ):
                L0Conf().setProp("ReplaceL0BanksWithEmulated",True)
            elif ( self.getProp("ReplayL0DUOnly") ):
                L0Conf().setProp("ReplayL0DU",True)
            else:
                L0Conf().setProp("SimulateL0",True)
        mainSeq.Members.append(l0seq)

        if self.getProp('Monitor'):
            from Configurables import L0DUReportMonitor
            l0du_mon = L0DUReportMonitor()
            mainSeq.Members.append(l0du_mon)

        if self.getProp('L0Filter'):
            from Configurables import LoKi__L0Filter
            self._setupLoKi()
            l0filter = LoKi__L0Filter('L0Filter', Code=self.getProp('L0Filter'))
            mainSeq.Members.append(l0filter)

        requirements = []
        if self.getProp('EventPreFilters') or self.getProp('L0Filter'):
            requirements = [mainSeq.name()]
        self._configureOutput(requirements)

        ApplicationMgr().TopAlg.append(mainSeq)

        # Use TimingAuditor for timing, suppress printout from SequencerTimerTool
        from Configurables import AuditorSvc,SequencerTimerTool
        ApplicationMgr().ExtSvc += [ 'ToolSvc', 'AuditorSvc' ]
        ApplicationMgr().AuditAlgorithms = True
        AuditorSvc().Auditors += [ 'TimingAuditor' ]
        if not SequencerTimerTool().isPropertySet( "OutputLevel" ):
            SequencerTimerTool().OutputLevel = 4
